#if !defined(SERIALCLASS_H_INCLUDED)
#define SERIALCLASS_H_INCLUDED

#include <windows.h>
#include <stdio.h>
#include <stdlib.h>

#define DEFAULT_COM        "\\\\.\\COM10" // "COM1"..."COM9"
#define ARDUINO_WAIT_TIME  0              // Time [ms] to wait for the arduino board to reset


class Serial
{
  public:
	// Singleton creator
    static Serial& GetReference ()
    {
        static Serial instance;
        return instance;
    }

	// Singleton creator
    static Serial* GetPtr () {
      return &GetReference();
    }

    // Check for available characters (RX)
    int Available ();

    // Read data in a buffer, if nbChar is greater than the
    // maximum number of bytes available, it will return only the
    // bytes available. The function return -1 when nothing could
    // be read, the number of bytes actually read.
    int ReadData (char *buffer, unsigned int nbChar);

    // Writes data from a buffer through the Serial connection
    // return true on success.
    bool WriteData (const unsigned char *buffer, unsigned int nbChar);

    // Check if we are actually connected
    bool IsConnected ();

    // Transmit (buffered) data to arduino
    bool Transmit ();


  private:
    // Private ctor
	// Initialize Serial communication with the given COM port
    Serial (const char *portName = DEFAULT_COM);

    // Private dtor.
	// Close the connection
    ~Serial ();

    HANDLE  hSerial;   // Serial comm (file-)handle
    bool    connected; // Connection status
    COMSTAT status;    // Various information about the connection
    DWORD   errors;    // Keeps track of last error

    // AGC channels and their 'mirror' values (for change detection)
	unsigned char channel_010[12][3];
    unsigned char channel_011[3];
    unsigned char channel_013[3];
    unsigned char _channel_010[12][3];
    unsigned char _channel_011[3];
    unsigned char _channel_013[3];
};

#endif // SERIALCLASS_H_INCLUDED